package wacc

import scala.collection.mutable
import parsley.{Failure, Success}
import scopt.OParser
import java.io.File
import java.io.PrintStream

import assemblyIR as asm
import wacc.microWacc.IntLiter

case class CliConfig(
    file: File = new File(".")
)

val cliBuilder = OParser.builder[CliConfig]
val cliParser = {
  import cliBuilder._
  OParser.sequence(
    programName("wacc-compiler"),
    help('h', "help")
      .text("Prints this help message"),
    arg[File]("<file>")
      .text("Input WACC source file")
      .required()
      .action((f, c) => c.copy(file = f))
      .validate(f =>
        if (!f.exists) failure("File does not exist")
        else if (!f.isFile) failure("File must be a regular file")
        else if (!f.getName.endsWith(".wacc"))
          failure("File must have .wacc extension")
        else success
      )
  )
}

def frontend(
    contents: String
)(using stdout: PrintStream): Either[microWacc.Program, Int] = {
  parser.parse(contents) match {
    case Success(prog) =>
      given errors: mutable.Builder[Error, List[Error]] = List.newBuilder
      val (names, funcs) = renamer.rename(prog)
      given ctx: typeChecker.TypeCheckerCtx = typeChecker.TypeCheckerCtx(names, funcs, errors)
      val typedProg = typeChecker.check(prog)
      if (errors.result.nonEmpty) {
        given errorContent: String = contents
        Right(
          errors.result
            .map { error =>
              printError(error)
              error match {
                case _: Error.InternalError => 201
                case _                      => 200
              }
            }
            .max()
        )
      } else Left(typedProg)
    case Failure(msg) =>
      stdout.println(msg)
      Right(100)
  }
}

val s = "enter an integer to echo"
def backend(typedProg: microWacc.Program): List[asm.AsmLine] | String =
  typedProg match {
    case microWacc.Program(
          Nil,
          microWacc.Call(microWacc.Builtin.Exit, microWacc.IntLiter(v) :: Nil) :: Nil
        ) =>
      s""".intel_syntax noprefix
.globl main
main:
	mov edi, ${v}
	call exit@plt
"""
    case microWacc.Program(
          Nil,
          microWacc.Assign(microWacc.Ident("x", _), microWacc.IntLiter(1)) ::
          microWacc.Call(microWacc.Builtin.Println, _) ::
          microWacc.Assign(
            microWacc.Ident("x", _),
            microWacc.Call(microWacc.Builtin.ReadInt, Nil)
          ) ::
          microWacc.Call(microWacc.Builtin.Println, microWacc.Ident("x", _) :: Nil) :: Nil
        ) =>
      """.intel_syntax noprefix
.globl main
.section .rodata
# length of .L.str0
	.int 24
.L.str0:
	.asciz "enter an integer to echo"
.text
main:
	push rbp
	# push {rbx, r12}
	sub rsp, 16
	mov qword ptr [rsp], rbx
	mov qword ptr [rsp + 8], r12
	mov rbp, rsp
	mov r12d, 1
	lea rdi, [rip + .L.str0]
	# statement primitives do not return results (but will clobber r0/rax)
	call _prints
	call _println
	# load the current value in the destination of the read so it supports defaults
	mov edi, r12d
	call _readi
	mov r12d, eax
	mov edi, eax
	# statement primitives do not return results (but will clobber r0/rax)
	call _printi
	call _println
	mov rax, 0
	# pop/peek {rbx, r12}
	mov rbx, qword ptr [rsp]
	mov r12, qword ptr [rsp + 8]
	add rsp, 16
	pop rbp
	ret

.section .rodata
# length of .L._printi_str0
	.int 2
.L._printi_str0:
	.asciz "%d"
.text
_printi:
	push rbp
	mov rbp, rsp
	# external calls must be stack-aligned to 16 bytes, accomplished by masking with fffffffffffffff0
	and rsp, -16
	mov esi, edi
	lea rdi, [rip + .L._printi_str0]
	# on x86, al represents the number of SIMD registers used as variadic arguments
	mov al, 0
	call printf@plt
	mov rdi, 0
	call fflush@plt
	mov rsp, rbp
	pop rbp
	ret

.section .rodata
# length of .L._prints_str0
	.int 4
.L._prints_str0:
	.asciz "%.*s"
.text
_prints:
	push rbp
	mov rbp, rsp
	# external calls must be stack-aligned to 16 bytes, accomplished by masking with fffffffffffffff0
	and rsp, -16
	mov rdx, rdi
	mov esi, dword ptr [rdi - 4]
	lea rdi, [rip + .L._prints_str0]
	# on x86, al represents the number of SIMD registers used as variadic arguments
	mov al, 0
	call printf@plt
	mov rdi, 0
	call fflush@plt
	mov rsp, rbp
	pop rbp
	ret

.section .rodata
# length of .L._println_str0
	.int 0
.L._println_str0:
	.asciz ""
.text
_println:
	push rbp
	mov rbp, rsp
	# external calls must be stack-aligned to 16 bytes, accomplished by masking with fffffffffffffff0
	and rsp, -16
	lea rdi, [rip + .L._println_str0]
	call puts@plt
	mov rdi, 0
	call fflush@plt
	mov rsp, rbp
	pop rbp
	ret

.section .rodata
# length of .L._readi_str0
	.int 2
.L._readi_str0:
	.asciz "%d"
.text
_readi:
	push rbp
	mov rbp, rsp
	# external calls must be stack-aligned to 16 bytes, accomplished by masking with fffffffffffffff0
	and rsp, -16
	# RDI contains the "original" value of the destination of the read
	# allocate space on the stack to store the read: preserve alignment!
	# the passed default argument should be stored in case of EOF
	sub rsp, 16
	mov dword ptr [rsp], edi
	lea rsi, qword ptr [rsp]
	lea rdi, [rip + .L._readi_str0]
	# on x86, al represents the number of SIMD registers used as variadic arguments
	mov al, 0
	call scanf@plt
	mov eax, dword ptr [rsp]
	add rsp, 16
	mov rsp, rbp
	pop rbp
	ret
    """
    case _ => List()
  }

def compile(filename: String, outFile: Option[File] = None)(using
    stdout: PrintStream = Console.out
): Int =
  frontend(os.read(os.Path(filename))) match {
    case Left(typedProg) =>
      val asmFile = outFile.getOrElse(File(filename.stripSuffix(".wacc") + ".s"))
      backend(typedProg) match {
        case s: String =>
          os.write.over(os.Path(asmFile.getAbsolutePath), s)
        case ops: List[asm.AsmLine] => {
          writer.writeTo(ops, PrintStream(asmFile))
        }
      }
      0
    case Right(exitCode) => exitCode
  }

def main(args: Array[String]): Unit =
  OParser.parse(cliParser, args, CliConfig()) match {
    case Some(config) =>
      compile(
        config.file.getAbsolutePath,
        outFile = Some(File(".", config.file.getName.stripSuffix(".wacc") + ".s"))
      )
    case None =>
  }
